unit ExceptionHandlingU;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls;

type
  TForm1 = class(TForm)
    GroupBox1: TGroupBox;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    GroupBox2: TGroupBox;
    Button4: TButton;
    Button5: TButton;
    Button6: TButton;
    Button7: TButton;
    dlgSave: TSaveDialog;
    Button8: TButton;
    Button9: TButton;
    Button10: TButton;
    dlgColour: TColorDialog;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Button6Click(Sender: TObject);
    procedure Button7Click(Sender: TObject);
    procedure Button8Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button9Click(Sender: TObject);
    procedure Button10Click(Sender: TObject);
  private
    IniName: String;
    procedure A;
    procedure B;
  public
    procedure AppExceptionHandler(Sender: TObject; E: Exception);
  end;

  ELXFError = class(Exception);
  ELXFMathError = class(ELXFError);
  ELXFZeroArg = class(ELXFError);
  ELXFTooFewArgs = class(ELXFError);

var
  Form1: TForm1;

implementation

{$R *.xfm}

uses
{$ifdef LINUX}
  Libc,
{$endif}
{$ifdef MSWINDOWS}
  Windows,
{$endif}
  IniFiles, Math;

const
  FormSettings = 'Form Settings';
  FormColour = 'Form Colour';

procedure WriteIniSetting(const IniFile, Section, Ident, Value: String); overload;
begin
  with TIniFile.Create(IniFile) do
    try
      WriteString(Section, Ident, Value);
      UpdateFile
    finally
      Free
    end
end;

procedure WriteIniSetting(const IniFile, Section, Ident: String; Value: Integer); overload;
begin
  WriteIniSetting(IniFile, Section, Ident, IntToStr(Value))
end;

function ReadIniSetting(const IniFile, Section, Ident, Value: String): String; overload;
begin
  with TIniFile.Create(IniFile) do
    try
      Result := ReadString(Section, Ident, Value)
    finally
      Free
    end
end;

function ReadIniSetting(const IniFile, Section, Ident: String; Value: Integer): Integer; overload;
begin
  Result := StrToInt(ReadIniSetting(IniFile, Section, Ident, IntToStr(Value)))
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  Tag := StrToInt(Application.ExeName)
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  (Sender as TForm).Hide
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  Lst: TListBox;
begin
  Lst := TListBox.Create(Self);
  Lst.Parent := Self;
  Lst.SendToBack;
  Lst.Items.LoadFromFile('Non-existentFile.txt');
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
  //SIGFPE signal
  Tag := 100 div Tag
end;

procedure TForm1.Button5Click(Sender: TObject);
var
  PI: PInteger;
begin
  PI := nil;
  //SIGSEGV signal
  Caption := IntToStr(PI^)
end;

procedure TForm1.AppExceptionHandler(Sender: TObject; E: Exception);
begin
  //Default response to unhandled exceptions
  //Application.ShowException(E)

  //Custom response to unhandled exceptions
  MessageDlg(
    Format('An %s occurred:'#10#10'"%s"', [E.ClassName, E.Message]),
    mtError, [mbOk], 0)
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Application.OnException := AppExceptionHandler;
  IniName := GetEnvironmentVariable('HOME') + '/.DynamicObjects';
  Color := ReadIniSetting(IniName, FormSettings, FormColour, Color);
end;

function CalcFloats(Floats: array of Double): Double;
var
  I: Integer;
begin
  if Length(Floats) < 2 then
    raise ELXFTooFewArgs.Create('Too few arguments');
  Result := 1;
  for I := Low(Floats) to High(Floats) do
  begin
    if Floats[I] = 0 then
      raise ELXFZeroArg.Create('Arguments must be non-zero');
    Result := Result * Floats[I]
  end
end;

procedure TForm1.Button6Click(Sender: TObject);
var
  Radius, Area, Volume: Double;
  RadiusStr: String;
begin
  RadiusStr := '1.0';
  if InputQuery('Trigonometry 101', 'Enter a radius', RadiusStr) then
    try
      Radius := StrToFloat(RadiusStr);
      Area := Pi * Sqr(Radius);
      Volume := Pi * Power(Radius, 3) * 4/3;
      ShowMessageFmt(
        'A circle with radius %g has area %g'#10 +
        'A sphere with radius %g has volume %g'#10 +
        'Calculation performed at %s on %s',
        [Radius, Area, Radius, Volume,
         TimeToStr(Time), DateToStr(Date)])
    except
      on EConvertError do
        MessageDlg(
          'Your entry was not recognised as a valid number',
          mtError, [mbOk], 0)
    end
end;

procedure TForm1.Button7Click(Sender: TObject);
var
  TF: TextFile;
begin
  if dlgSave.Execute then
    try
      AssignFile(TF, dlgSave.FileName);
      try
        Append(TF);
        WriteLn(TF, 'Hello world');
      finally
        CloseFile(TF)
      end
    except
      on E: EInOutError do
        case E.ErrorCode of
        {$ifdef LINUX}ENOENT{$endif}{$ifdef MSWINDOWS}ERROR_FILE_NOT_FOUND{$endif}:
           ShowMessage('"%s" cannot be located', [dlgSave.FileName]);
        {$ifdef LINUX}EACCES{$endif}{$ifdef MSWINDOWS}ERROR_ACCESS_DENIED{$endif}:
           ShowMessage('You do not have permission to overwrite this file');
        else
          if E.ErrorCode < 100 then
            ShowMessage('I/O Error'#10#9'Error code: %d'#10#9 +
              'Error msg: %s', [E.ErrorCode, SysErrorMessage(E.ErrorCode)])
          else
            ShowMessage('I/O Error'#10#9'Error code: %d'#10#9 +
              'Error msg: %s', [E.ErrorCode, E.Message]);
        end;
    end
end;

procedure TForm1.Button8Click(Sender: TObject);
begin
  //more code
  A
  //more code
end;

procedure TForm1.A;
begin
  try
    //more code
    B
    //more code
  except
    on E: EConvertError do
    begin
      ShowMessage('We just logged a problem:'#10#10 +
        'A %s exception occurred, saying:'#10'"%s"',
      [E.ClassName, E.Message])
    end
  end
end;

procedure TForm1.B;
var
  Radius, Area: Double;
  RadiusStr: String;
begin
  RadiusStr := '1.0';
  if InputQuery('Trigonometry 101', 'Enter a radius', RadiusStr) then
  begin
    Radius := StrToFloat(RadiusStr);
    Area := Pi * Sqr(Radius);
    ShowMessageFmt('A circle with radius %g has area %g', [Radius, Area])
  end
end;

procedure TForm1.Button9Click(Sender: TObject);
var
  Radius, Res: Double;
begin
  Radius := 10; //this should be read from the user
  Res := CalcFloats([Pi, Radius, Radius]);
  ShowMessage('The answer was %g', [Res])
end;

procedure TForm1.Button10Click(Sender: TObject);
begin
  dlgColour.Color := Color; //initialise dialog to show form's colour
  if dlgColour.Execute then
  begin
    Color := dlgColour.Color;
    WriteIniSetting(IniName, FormSettings, FormColour, Color);
  end
end;

end.
